import os
import boto3
import botocore
import logging
import base64

from mind_castle.secret_store_base import SecretStoreBase, MindCastleSecret
from mind_castle.exceptions import (
    RetrieveSecretException,
    CreateSecretException,
)

logger = logging.getLogger(__name__)


class AWSKMSSecretStore(SecretStoreBase):
    """
    Uses AWS KMS to encrypt secrets.
    """

    store_type = "awskms"
    required_config = [
        [
            "MIND_CASTLE_AWS_REGION",
            "MIND_CASTLE_AWS_KMS_KEY_ID",
            "MIND_CASTLE_AWS_SECRET_ACCESS_KEY",
            "MIND_CASTLE_AWS_KMS_KEY_ARN",
        ],
        [
            "MIND_CASTLE_AWS_REGION",
            "MIND_CASTLE_AWS_USE_ENV_AUTH",
            "MIND_CASTLE_AWS_KMS_KEY_ARN",
        ],
    ]
    optional_config = []

    KMS_ENCRYPTION_ALGORITHM = "RSAES_OAEP_SHA_256"

    def __init__(self):
        if all([os.environ.get(req) for req in self.required_config[0]]):
            # Configure with secret key
            self.client = boto3.client(
                "kms",
                region_name=os.environ["MIND_CASTLE_AWS_REGION"],
                aws_access_key_id=os.environ["MIND_CASTLE_AWS_ACCESS_KEY_ID"],
                aws_secret_access_key=os.environ["MIND_CASTLE_AWS_SECRET_ACCESS_KEY"],
            )
        else:
            # Assume the environment is configured with the correct credentials
            self.client = boto3.client(
                "kms", region_name=os.environ["MIND_CASTLE_AWS_REGION"]
            )
        self.kms_key_arn = os.environ["MIND_CASTLE_AWS_KMS_KEY_ARN"]

    def retrieve_secret(self, secret: MindCastleSecret) -> str:
        encrypted_value = secret.encrypted_value
        try:
            response = self.client.decrypt(
                KeyId=self.kms_key_arn,
                EncryptionAlgorithm=self.KMS_ENCRYPTION_ALGORITHM,
                CiphertextBlob=base64.b64decode(encrypted_value),
            )
        except botocore.exceptions.ClientError as e:
            logger.exception(f"Error retrieving secret {secret.key}:")
            raise RetrieveSecretException(e, encrypted_value)

        return response.get("Plaintext").decode("utf-8")

    def create_secret(self, value: str) -> MindCastleSecret:
        try:
            # Response is always a success. This throws an exception if it fails
            response = self.client.encrypt(
                KeyId=self.kms_key_arn,
                EncryptionAlgorithm=self.KMS_ENCRYPTION_ALGORITHM,
                Plaintext=value.encode("utf-8"),
            )
        except botocore.exceptions.ClientError as e:
            raise CreateSecretException(e)

        # ARN is not needed but returned just in case. we could also add tags etc to this response
        return MindCastleSecret(
            mind_castle_secret_type=self.store_type,
            encrypted_value=base64.b64encode(response.get("CiphertextBlob")).decode(
                "utf-8"
            ),
        )

    def update_secret(self, secret: MindCastleSecret, value: str) -> MindCastleSecret:
        new_secret = self.create_secret(value)
        new_secret.key = secret.key
        return new_secret

    def delete_secret(self, secret: MindCastleSecret) -> None:
        pass



from moto import mock_aws
import pytest
import boto3
import os
import base64

from mind_castle.stores.aws_kms import AWSKMSSecretStore


@pytest.fixture(scope="module")
def mockaws():
    with mock_aws():
        kms = boto3.client("kms", region_name="us-east-1")
        response = kms.create_key(
            Description="Test Key", KeyUsage="ENCRYPT_DECRYPT", KeySpec="RSA_4096"
        )
        os.environ["MIND_CASTLE_AWS_KMS_KEY_ARN"] = response["KeyMetadata"]["Arn"]
        yield


def test_create_secret(mockaws):
    secret_store = AWSKMSSecretStore()
    response = secret_store.create_secret("some_secret_value")

    client = secret_store.client  # Use the same auth etc as the store
    boto_response = client.decrypt(
        CiphertextBlob=base64.b64decode(response.encrypted_value),
        KeyId=os.environ["MIND_CASTLE_AWS_KMS_KEY_ARN"],
    )
    assert boto_response["Plaintext"].decode("utf-8") == "some_secret_value"
    assert response.mind_castle_secret_type == secret_store.store_type


# TODO: Add tests for bad-path. E.g. invalid key arn, invalid ciphertext


def test_retrieve_secret(mockaws):
    secret_store = AWSKMSSecretStore()
    response = secret_store.create_secret("some_retrievable_value")

    assert secret_store.retrieve_secret(response) == "some_retrievable_value"


def test_delete_secret():
    secret_store = AWSKMSSecretStore()
    secret_store.delete_secret(
        {"mind_castle_secret_type": "awskms", "key": "some_deletable_key"}
    )
    # Just don't throw an error to pass


def test_update_secret():
    secret_store = AWSKMSSecretStore()
    create_response = secret_store.create_secret("some_updatable_value")

    update_response = secret_store.update_secret(create_response, "some_updated_value")

    newvalue = secret_store.retrieve_secret(update_response)
    assert newvalue == "some_updated_value"
